/*
 * Copyright (c) 2020 Fingerprint Cards AB
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file    ftdi_spi.c
 * @brief   File to handle spi communication
 */

#include "ftdi_spi.h"
#include <ftd2xx.h>
#include <stdio.h>
#include <stdlib.h>
#include "platform.h"

#define FTDI_ADBUS0_CLK                      (1 << 0)
#define FTDI_ADBUS1_MOSI                     (1 << 1)
#define FTDI_ADBUS2_MISO                     (1 << 2)
#define FTDI_ADBUS3_CS                       (1 << 3)
#define FTDI_ADBUS4_GPIOL0                   (1 << 4)
#define FTDI_ADBUS5_GPIOL1                   (1 << 5)
#define FTDI_ADBUS6_GPIOL2                   (1 << 6)
#define FTDI_ADBUS7_GPIOL3                   (1 << 7)

#define FTDI_ACBUS0_GREEN                    (1 << 0)
#define FTDI_ACBUS1_RED                      (1 << 1)
#define FTDI_ACBUS2_YELLOW                   (1 << 2)
#define FTDI_ACBUS3_GPIOH3                   (1 << 3)
#define FTDI_ACBUS4_GPIOH4                   (1 << 4)
#define FTDI_ACBUS5_IRQ                      (1 << 5)
#define FTDI_ACBUS6_RESET                    (1 << 6)
#define FTDI_ACBUS7_GPIOH7                   (1 << 7)

#define FTDI_DISABLE_CLOCKDIV_5              0x8A
#define FTDI_ENABLE_CLOCKDIV_5               0x8B
#define FTDI_ENABLE_3_PHASE                  0x8C
#define FTDI_DISABLE_3_PHASE                 0x8D
#define FTDI_DISABLE_ADAPTIVE_CLOCKING       0x97
#define FTDI_SET_LOW_PINS                    0x80
#define FTDI_SET_HIGH_PINS                   0x82
#define FTDI_READ_LOW_PINS                   0x81
#define FTDI_READ_HIGH_PINS                  0x83
#define FTDI_SET_CLOCK_DIV                   0x86
#define FTDI_DISABLE_LOOPBACK                0x85
#define FTDI_SEND_IMMEDIATE                  0x87

#define FTDI_OUT_NEG                         0x11
#define FTDI_IN_POS                          0x20

#define FTDI_TRANSFER_SIZE_L(x)              (((x) - 1) & 0xFF)
#define FTDI_TRANSFER_SIZE_H(x)              ((((x) - 1) >> 8) & 0xFF)

//  60 MHz / (2 * (x+1))
#define FTDI_SPI_CLK_30_MHZ                  0
#define FTDI_SPI_CLK_15_MHZ                  1
#define FTDI_SPI_CLK_10_MHZ                  2
#define FTDI_SPI_CLK_7_5_MHZ                 3
#define FTDI_SPI_CLK_6_MHZ                   4
#define FTDI_SPI_CLK_5_MHZ                   5
#define FTDI_SPI_CLK_4_29_MHZ                6
#define FTDI_SPI_CLK_3_75_MHZ                7
#define FTDI_SPI_CLK_3_33_MHZ                8
#define FTDI_SPI_CLK_3_MHZ                   9
#define FTDI_SPI_CLK_2_7_MHZ                 10
#define FTDI_SPI_CLK_2_5_MHZ                 11
#define FTDI_SPI_CLK_2_MHZ                   14
#define FTDI_SPI_CLK_1_5_MHZ                 19
#define FTDI_SPI_CLK_1_MHZ                   29
#define FTDI_SPI_CLK_0_5_MHZ                 59

static const int spi_speed[] =
{
        FTDI_SPI_CLK_0_5_MHZ,  500,
        FTDI_SPI_CLK_1_MHZ,    1000,
        FTDI_SPI_CLK_1_5_MHZ,  1500,
        FTDI_SPI_CLK_2_MHZ,    2000,
        FTDI_SPI_CLK_2_5_MHZ,  2500,
        FTDI_SPI_CLK_2_7_MHZ,  2700,
        FTDI_SPI_CLK_3_MHZ,    3000,
        FTDI_SPI_CLK_3_33_MHZ, 3330,
        FTDI_SPI_CLK_3_75_MHZ, 3750,
        FTDI_SPI_CLK_4_29_MHZ, 4290,
        FTDI_SPI_CLK_5_MHZ,    5000,
        FTDI_SPI_CLK_6_MHZ,    6000,
        FTDI_SPI_CLK_7_5_MHZ,  7500,
        FTDI_SPI_CLK_10_MHZ,   10000,
        FTDI_SPI_CLK_15_MHZ,   15000,
        FTDI_SPI_CLK_30_MHZ,   30000,
};

FT_HANDLE ftHandle = NULL; // Handle of the FTDI device

int get_valid_baudrate(int *baudrate)
{
    int s = spi_speed[0];
    int speed = spi_speed[1];

    for (int i = 2; i < sizeof(spi_speed)/sizeof(spi_speed[0]); i += 2) {
        if(spi_speed[i + 1] > *baudrate) {
            break;
        }
        s = spi_speed[i];
        speed = spi_speed[i + 1];
    }
    *baudrate = speed;
    return s;
}

uint32_t ftdi_spi_open(int *baudrate)
{
    FT_STATUS ftStatus;

    ftStatus = FT_Open(0, &ftHandle);

    if (ftStatus != FT_OK) {
        ftHandle = NULL;
        return 1;
    }

    DWORD uint8_ts_in_buf = 0;
    FT_ResetDevice(ftHandle);


    FT_GetQueueStatus(ftHandle, &uint8_ts_in_buf);

    if (uint8_ts_in_buf > 0) {
        DWORD read;
        BYTE *buff = (BYTE*)malloc(uint8_ts_in_buf);
        if (FT_Read(ftHandle, buff, uint8_ts_in_buf, &read)) {
            free(buff);
            return 2;
        }
        free(buff);
    }

    FT_SetChars(ftHandle, 0, 0, 0, 0);
    FT_SetLatencyTimer(ftHandle, 32);

    FT_STATUS ret;

    ret = FT_SetBitMode(ftHandle, 0x00, FT_BITMODE_MPSSE);

    if (ret != FT_OK) {
        return 3;
    }

    BYTE setupMsg[] = {
        FTDI_DISABLE_ADAPTIVE_CLOCKING,
        FTDI_DISABLE_LOOPBACK,
        FTDI_DISABLE_CLOCKDIV_5,
        FTDI_DISABLE_3_PHASE,
        FTDI_SET_CLOCK_DIV,
        get_valid_baudrate(baudrate),
        0,
        FTDI_SET_LOW_PINS,
        FTDI_ADBUS3_CS,
        FTDI_ADBUS1_MOSI | FTDI_ADBUS0_CLK | FTDI_ADBUS3_CS,
        FTDI_SET_HIGH_PINS,
        FTDI_ACBUS6_RESET,
        FTDI_ACBUS6_RESET | FTDI_ACBUS0_GREEN | FTDI_ACBUS1_RED | FTDI_ACBUS2_YELLOW,

    };

    DWORD written = 0;

    ret = FT_Write(ftHandle, setupMsg, sizeof(setupMsg), &written);

    if (ret != FT_OK) {
        return 4;
    }

    return 0;
}

uint32_t ftdi_spi_write(const uint8_t* buffer, int length)
{
    FT_STATUS ret;
    int mpsseLen = length + 9;
    DWORD written = 0;
    int p = 0;

    if (!buffer || length < 0 || !ftHandle)
        return 1;

    uint8_t *mpsse = (uint8_t*)malloc(mpsseLen);

    if (!mpsse)
        return 1;

    ftdi_spi_set_cs(0);

    // SPI Send
    mpsse[p++] = FTDI_OUT_NEG;
    mpsse[p++] = FTDI_TRANSFER_SIZE_L(length);
    mpsse[p++] = FTDI_TRANSFER_SIZE_H(length);

    // Queue Data
    for (int i = 0; i < length; i++)
    {
        mpsse[p++] = buffer[i];
    }

    // CS
    mpsse[p++] = FTDI_SET_LOW_PINS;
    mpsse[p++] = FTDI_ADBUS3_CS;
    mpsse[p++] = FTDI_ADBUS0_CLK + FTDI_ADBUS3_CS + FTDI_ADBUS1_MOSI;

    ret = FT_Write(ftHandle, mpsse, p, &written);

    if (written != p) {
        WRITE("Failed to write command");
    }

    free(mpsse);

    return (ret == FT_OK) ? 0 : 1;
}

uint32_t ftdi_spi_read(uint8_t * buffer, int length)
{
    FT_STATUS ret;
    DWORD written;
    int p = 0;
    int mpsseLen = length + 13;
    int irq = 0;

    if (!length || length < 0 || !ftHandle)
        return 1;

    while(!irq)
    {
        if (ftdi_spi_check_irq(&irq))
            return 1;
    }

    uint8_t *mpsse = (uint8_t*)malloc(mpsseLen);
    if (!mpsse)
        return 1;

    // CS
    mpsse[p++] = FTDI_SET_LOW_PINS;
    mpsse[p++] = 0;
    mpsse[p++] = FTDI_ADBUS0_CLK + FTDI_ADBUS3_CS + FTDI_ADBUS1_MOSI;

    // SPI Read
    mpsse[p++] = FTDI_IN_POS;
    mpsse[p++] = FTDI_TRANSFER_SIZE_L(length);
    mpsse[p++] = FTDI_TRANSFER_SIZE_H(length);

    // CS
    mpsse[p++] = FTDI_SET_LOW_PINS;
    mpsse[p++] = FTDI_ADBUS3_CS;
    mpsse[p++] = FTDI_ADBUS0_CLK + FTDI_ADBUS3_CS + FTDI_ADBUS1_MOSI;

    // Don't hold on to data
    mpsse[p++] = FTDI_SEND_IMMEDIATE;
    FT_ClrRts(ftHandle);

    ret = FT_Write(ftHandle, mpsse, p, &written);

    free(mpsse);

    if (ret != FT_OK)
        return 1;

    ret = FT_Read(ftHandle, buffer, length, &written);
    FT_ClrRts(ftHandle);

    return (ret == FT_OK) ? 0 : 1;
}

uint32_t ftdi_spi_set_cs(int cs)
{
    FT_STATUS ret;
    DWORD written;

    uint8_t mpsse[3];

    if (!ftHandle)
        return 1;

    // CS
    mpsse[0] = FTDI_SET_LOW_PINS;
    mpsse[1] = cs ? FTDI_ADBUS3_CS : 0;
    mpsse[2] = FTDI_ADBUS0_CLK | FTDI_ADBUS3_CS | FTDI_ADBUS1_MOSI;

    ret = FT_Write(ftHandle, mpsse, 3, &written);

    return !(ret == FT_OK);
}

uint32_t ftdi_spi_check_irq(int *irq)
{
    FT_STATUS ret;
    DWORD written;
    uint8_t mpsse[1];
    uint8_t buffer[1];

    if (!ftHandle || !irq)
        return 1;

    *irq = 0;
    // Read interupt
    mpsse[0] = FTDI_READ_HIGH_PINS;

    ret = FT_Write(ftHandle, mpsse, 1, &written);
    if(ret != FT_OK) {
        return 1;
    }
    ret = FT_Read(ftHandle, buffer, 1, &written);
    if(ret != FT_OK) {
        return 1;
    }

    if(buffer[0] & FTDI_ACBUS5_IRQ)
    {
        *irq = 1;
    }
    return 0;
}

uint32_t ftdi_spi_close(void)
{
    if (!ftHandle)
        return 1;

    return !(FT_OK == FT_Close(&ftHandle));
}

uint32_t ftdi_set_reset_level(int level)
{
    FT_STATUS ret;
    DWORD written;

    uint8_t mpsse[3];

    if (!ftHandle)
        return 1;

    mpsse[0] = FTDI_SET_HIGH_PINS;
    mpsse[1] = (level == 0) ? 0 : FTDI_ACBUS6_RESET;
    mpsse[2] = FTDI_ACBUS6_RESET | FTDI_ACBUS0_GREEN | FTDI_ACBUS1_RED | FTDI_ACBUS2_YELLOW;

    ret = FT_Write(ftHandle, mpsse, 3, &written);

    return (ret == FT_OK && written == 3) ? 0 : 1;
}
