/*
 * Copyright (c) 2020 Fingerprint Cards AB
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file    platform
 * @brief
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include "ftdi_spi.h"
#include "fpc_com_result.h"
#include "platform.h"

bool platform_com_init(char *port, int *baudrate, int timeout)
{
    uint32_t status = ftdi_spi_open(baudrate);
    if (status) {
        WRITE("FTDI open failed with status %d\r\n", status);
#if !defined(_WIN32) && (defined(__unix__) || defined(__unix)) && !defined(__CYGWIN__)
        WRITE("Try sudo rmmod ftdi_sio and sudo rmmod usb_serial");
#endif
        return false;
    }

    return true;
}

fpc_com_result_t platform_com_send(uint16_t size, const uint8_t *data, uint32_t timeout,
        void *session)
{
    fpc_com_result_t res = FPC_COM_RESULT_OK;

    uint32_t status = ftdi_spi_write(data, size);
    if (status) {
        WRITE("FTDI send failed with status %d\r\n", status);
        res = FPC_COM_RESULT_IO_ERROR;
    }

    return res;
}

fpc_com_result_t platform_com_receive(uint16_t size, uint8_t *data, uint32_t timeout,
        void *session)
{
    fpc_com_result_t res = FPC_COM_RESULT_OK;

    uint32_t status = ftdi_spi_read(data,size);
    if (status) {
        WRITE("FTDI read failed with status %d\r\n", status);
        res = FPC_COM_RESULT_IO_ERROR;
    }

    return res;
}

uint64_t platform_get_time(void)
{
#ifdef _WIN32
    // GetTickCount64() not available in mingw.
    return GetTickCount();
#else
    struct timeval current_time;
    uint64_t time_in_ms;

    gettimeofday(&current_time, NULL);
    /* Seconds and microseconds are converted to milliseconds */
    time_in_ms = current_time.tv_usec / 1000 + current_time.tv_sec * 1000;

    return time_in_ms;
#endif
}

void platform_delay_us(uint64_t us)
{
#ifdef _WIN32
    if (us < 1000) {
        /* Windows has no sleep function with high accuracy. Use busy wait for short
        delays and Sleep for longer */
        LARGE_INTEGER now, end, freq;
        QueryPerformanceFrequency(&freq);
        QueryPerformanceCounter(&now);
        end.QuadPart = now.QuadPart + us * freq.QuadPart / 1000000;
        while (end.QuadPart > now.QuadPart) {
            QueryPerformanceCounter(&now);
        }
    }
    else {
        Sleep((DWORD)(us / 1000));
    }
#else
    usleep(us);
#endif
}

void platform_clear_screen(void)
{
#ifdef _WIN32
    system("cls");
#else
    system("clear");
#endif
}

bool platform_check_irq(void)
{
    int irq;

    (void)ftdi_spi_check_irq(&irq);

    return !(irq == 0);
}

void platform_reset(void)
{
    ftdi_set_reset_level(0);
    platform_delay_us(1000);
    ftdi_set_reset_level(1);
}
